# Copyright (c) 2020-2021 Thomas Paviot (tpaviot@gmail.com)
#
# This file is part of ProcessScheduler.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

import processscheduler.jupyter_ui as jupui
from ipywidgets import VBox


class TestJupyerUI(unittest.TestCase):
    def test_load_jupyter_ui(self) -> None:
        self.assertTrue(isinstance(jupui.task_ui, VBox))
        self.assertTrue(isinstance(jupui.problem_ui, VBox))


if __name__ == "__main__":
    unittest.main()
